#include "mtest.h"

/*
 * This is a best-fit allocation engine. Slow but trivial.
 * The variables are still called ff and fitem, to ease diff.
 */

struct ff_item {
	struct ff_item *prev;
	struct ff_item *next;
	unsigned long size; /* payload size */
	unsigned long flags; /* also for data alignment */
	char d[0];
};

#define FF_FREE  0x1 /* flags */
#define DATA_OFFSET sizeof(struct ff_item)

static struct ff_item head; /* first item, with zero size and used */
static struct ff_item *loop; /* the start of the search loop */

/* first-fit find a hole */
static struct ff_item *ff_find(int size)
{
	struct ff_item *p = loop;
	struct ff_item *v = NULL; /* victim */
	int best = (1<<30);

	if (!p) return NULL;

	do {
		if (!(p->flags & FF_FREE) || p->size < size)
			goto again;
		if (p->size == size) {
			loop = p->next;
			return p;
		}
		if (p->size < best) {
			best = p->size;
			v = p;
		}
	again:
		p = p->next;
	} while (p != loop);

	return v;
}

/* Split a free block, leaving a free tail: size must be aligned */
static void ff_split(struct ff_item *fi, int size)
{
	struct ff_item *new;
	int nsize;

	new = (typeof(new))(fi->d + size);
	nsize = fi->size - size - DATA_OFFSET;
	fi->size = size;

	/* So, now "fi" is allocated, while "new" is new and free */
	new->size = nsize;
	new->prev = fi;
	new->next = fi->next;
	new->next->prev = new;
	fi->next = new;

	fi->flags &= ~FF_FREE;
	new->flags = FF_FREE;
}

/* Merge two consecutive free blocks, already out of tree */
static void ff_merge(struct ff_item *fi1, struct ff_item *fi2)
{
	/* we are sure fi1->prev and fi2->next are not free */
	fi1->size += fi2->size + DATA_OFFSET;
	fi1->next = fi2->next;
	fi1->next->prev = fi1;
}

/* The public functions are called dlmalloc and dlfree to ease linking */
void* dlmalloc(int size)
{
	struct ff_item *fi;
	int ns; /* new size */
	int allocsize;

	ns = M_ALIGN(size);
	allocsize = ns + sizeof(*fi);

	fi = ff_find(ns);
	if (fi) {
		/* The area is still marked as free. Split it */
		if (fi->size - ns > DATA_OFFSET)
			ff_split(fi, ns);
		fi->flags = 0;
		return fi->d;
	}
	/* Not found, request memory to the system */
	if (!head.next) {
		/* first area ever */
		fi = sbrk(allocsize);
		head.next = head.prev = fi;
		fi->next = fi->prev = &head;
		goto found;
	}
	if (head.prev->flags & FF_FREE)
		allocsize -= head.prev->size + DATA_OFFSET;
	fi = sbrk(allocsize);
	if (fi == (typeof(fi))-1)
		return NULL;
	/* If we extended last free block, don't link */
	if (head.prev->flags & FF_FREE) {
		fi = head.prev;
	} else {
		fi->prev = head.prev;
		fi->next = &head;
		fi->prev->next = fi;
		fi->next->prev = fi;
	}
found:
	fi->size = ns;
	fi->flags = 0;
	return fi->d;
}

void dlfree(void *area)
{
	struct ff_item *fi, *next, *prev;

	fi = (typeof(fi))(area - DATA_OFFSET);
	fi->flags = FF_FREE;
	prev = fi->prev;
	next = fi->next;

	if (!loop) {
		loop = fi; /* first free ever */
		return;
	}

	/* merge with next and prev. Note that head is never free. */
	if (next->flags & FF_FREE) {
		ff_merge(fi, next);
		if (loop == next) loop = &head;
	}
	if (prev->flags & FF_FREE ) {
		/* merge */
		ff_merge(prev, fi);
		if (loop == fi) loop = &head;
		fi = prev;
	}
}


/* Debugging tool */
#include <stdio.h>
void rbm_show(FILE *f, void *min)
{
	struct ff_item *h = &head;

	fprintf(f, "<<<<< head %p\n", h);

	while (h = h->next, h && h != &head) {
		fprintf(f, "%08x (free %08lx) %c (sz %4x)"
			" prev %08x   next %08x\n",
			(void *)h-min, (long)h + 16 - (long)min,
			h->flags ? 'F' : '-', h->size,
			(void *)h->prev-min, (void *)h->next-min);
	}

	fprintf(f, ">>>>>\n");
}
