#include "mtest.h"
#include "linux_rbtree.h"

/*
 * The idea here is simple: keep an rbtree of holes: on free we add to
 * the rbtree and on alloc we do best-fit or worst-fit.
 */

struct rbm_item {
	struct rbm_item *prev;
	struct rbm_item *next;
	unsigned long size; /* payload size */
	unsigned long flags; /* also for data alignment */
	union {
		char d[0];
		struct rb_node rb;
	} pl; /* payload */
};

#define RBM_FREE  0x1 /* flags */
#define DATA_OFFSET offsetof(struct rbm_item, pl)
#define MIN_ALLOC (sizeof(struct rbm_item) - DATA_OFFSET)

static struct rb_root rbm_tree = RB_ROOT;  /* tree of free areas */
static struct rbm_item head; /* list head: prev is last, next is first */

#if 0
/* worst-fit using rb_last */
static struct rbm_item *rbm_find(int size)
{
	struct rb_node *n = rb_last(&rbm_tree);
	struct rbm_item *ri = container_of(n, struct rbm_item, pl.rb);
	if (!n || ri->size < size)
		return NULL;
	rb_erase(&ri->pl.rb, &rbm_tree);
	return ri;
}
#else
/* better-fit (not best) */
static struct rbm_item *rbm_find(int size)
{
	struct rb_node **p = &rbm_tree.rb_node;
	struct rb_node *parent = NULL;
	struct rbm_item *ri, *victim = NULL;;

	/*
	 * If this is smaller, go bigger.
	 * If this is bigger, mark as victim and go smaller
	 */
	while (*p) {
		parent = *p;
		ri = rb_entry(parent, struct rbm_item, pl.rb);

		if (ri->size == size) {
			victim = ri;
			break;
		}
		if (ri->size < size) {
			p = &(*p)->rb_right;
		} else {
			victim = ri;
			p = &(*p)->rb_left;
		}
	}
	if (!victim)
		return NULL;
	rb_erase(&victim->pl.rb, &rbm_tree);
	return victim;
}
#endif

/* Insert a free item in the tree. list links must already be correct */
static void rbm_free(struct rbm_item *ri)
{
	struct rb_node **p = &rbm_tree.rb_node;
	struct rb_node *parent = NULL;
	struct rbm_item *other;

	ri->flags = RBM_FREE;
	/* Note: merging has already happened in the caller */
	while (*p) {
		parent = *p;
		other = rb_entry(parent, struct rbm_item, pl.rb);

		if (ri->size < other->size)
			p = &(*p)->rb_left;
		else
			p = &(*p)->rb_right;
	}
	rb_link_node(&ri->pl.rb, parent, p);
	rb_insert_color(&ri->pl.rb, &rbm_tree);
}

/* Split a free block, leaving a free tail: size must be aligned */
static void rbm_split(struct rbm_item *ri, int size)
{
	struct rbm_item *new;
	int nsize;

	new = (typeof(new))(ri->pl.d + size);
	nsize = ri->size - size - DATA_OFFSET;
	ri->size = size;

	/* So, now "ri" is allocated, while "new" is new and free */
	new->size = nsize;
	new->prev = ri;
	new->next = ri->next;
	new->next->prev = new;
	ri->next = new;

	ri->flags &= ~RBM_FREE;
	rbm_free(new);
}

/* Merge two consecutive free blocks, already out of tree */
static void rbm_merge(struct rbm_item *ri1, struct rbm_item *ri2)
{
	/* we are sure ri1->prev and ri2->next are not free */
	ri1->size += ri2->size + DATA_OFFSET;
	ri1->next = ri2->next;
	ri1->next->prev = ri1;
}

/* Return an item, removed from the tree and linked in thelist */
static struct rbm_item *rbm_alloc(int size)
{
	struct rbm_item *ri;
	int ns; /* new size */
	int allocsize;

	if (size < MIN_ALLOC)
		size = MIN_ALLOC;
	ns = M_ALIGN(size);
	allocsize = ns + DATA_OFFSET;
	ri = rbm_find(ns);
	if (ri) {
		/* The area is still marked as free. Split it */
		if (ri->size - ns >= DATA_OFFSET + MIN_ALLOC)
			rbm_split(ri, ns);
		ri->flags = 0;
		return ri;
	}
	/* Not found, request memory to the system */
	if (!head.next) {
		/* first area ever */
		ri = sbrk(allocsize);
		head.next = head.prev = ri;
		ri->next = ri->prev = &head;
		goto found;
	}
	if (head.prev->flags & RBM_FREE)
		allocsize -= head.prev->size + DATA_OFFSET;
	ri = sbrk(allocsize);
	if (ri == (typeof(ri))-1)
		return NULL;

	/* If we extended last free block, don't link */
	if (head.prev->flags & RBM_FREE) {
		ri = head.prev;
		rb_erase(&ri->pl.rb, &rbm_tree);
	} else {
		ri->prev = head.prev;
		ri->next = &head;
		ri->prev->next = ri;
		ri->next->prev = ri;
	}
 found:
	ri->size = ns;
	ri->flags = 0;
	return ri;
}

/* The public functions are called dlmalloc and dlfree to ease linking */
void* dlmalloc(int size)
{
	struct rbm_item *ri = rbm_alloc(size);
	if (!ri) return NULL;
	return ri->pl.d;
}

void dlfree(void *area)
{
	struct rbm_item *ri, *next, *prev;

	ri = (typeof(ri))(area - DATA_OFFSET);
	ri->flags = RBM_FREE;
	prev = ri->prev;
	next = ri->next;

	/* merge with next and prev. Note that head is never free. */
	if (next->flags & RBM_FREE) {
		rb_erase(&next->pl.rb, &rbm_tree);
		rbm_merge(ri, next);
	}
	if (prev->flags & RBM_FREE ) {
		/* merge */
		rb_erase(&prev->pl.rb, &rbm_tree);
		rbm_merge(prev, ri);
		ri = prev;
	}

	/* finally, put it in the tree */
	rbm_free(ri);
}


/* Debugging tool */
#include <stdio.h>
void rbm_show(FILE *f, void *min)
{
	struct rbm_item *h = &head;
	struct rbm_item *ri;
        struct rb_node *node;

	fprintf(f, "<<<<< head %p\n", h);

	while (h = h->next, h && h != &head) {
		fprintf(f, "%08x (free %08lx) %c (sz %4x)"
			" prev %08x   next %08x\n",
			(void *)h-min, (long)h + 16 - (long)min,
			h->flags ? 'F' : '-', h->size,
			(void *)h->prev-min, (void *)h->next-min);
	}

	for (node = rb_first(&rbm_tree); node; node = rb_next(node)) {
		ri = rb_entry(node, struct rbm_item, pl.rb);
		fprintf(f, " tree: %08x - size 0x%4x %6i\n",
			(void *)ri-min, ri->size, ri->size);
	}
	fprintf(f, ">>>>>\n");
}
